import asyncio
import os
import pkgutil
import sys
from ssl import SSLContext
from typing import Optional, List
import weakref

import aiohttp

from aiohttp import web, WSCloseCode

from .system import Page, UNDEFINED, Route

from .tools import replace_all, format_tag
from ._renderer import package as renderer


class Server:
    """
    Dazzler server

    :Routes:

        - ``/dazzler/update``
        - ``/dazzler/requirements/``

    """
    runner: web.AppRunner
    loop: asyncio.AbstractEventLoop
    app: web.Application
    site: Optional[web.TCPSite]

    def __init__(self, dazzler, loop=None, app: web.Application = None):
        """
        :param dazzler: Dazzler instance.
        :type dazzler: dazzler.Dazzler
        :param loop: Asyncio event loop.
        :param app: An existing aiohttp web application to use.
        """
        self.dazzler = dazzler
        self.loop = loop or asyncio.get_event_loop()
        self.app = app or web.Application()
        self.index = pkgutil.get_data(
            'dazzler', os.path.join('assets', 'index.html')
        ).decode()
        self.logger = self.dazzler.logger
        self.websockets = weakref.WeakSet()
        self.debug = False
        self.site = None

    def setup_routes(self, routes: List[Route] = None, debug: bool = False):
        """
        Setup routes for dazzler.

        :param routes: Routes to add the server.
        :param debug: Set debug mode.
        :return:
        """
        routes = routes or []
        prefix = self.dazzler.config.route_prefix
        self.debug = debug

        # Dazzler api.
        self.app.add_routes([
            web.get(
                f'{prefix}/dazzler/link',
                self._apply_middleware(self.route_get_page)
            )
        ] + [
            x.method.get_method()(
                x.path, self._apply_middleware(x.handler), name=x.name
            )
            for x in routes
        ])

        # User static directory.
        static_dir = os.path.join(
            self.dazzler.root_path, self.dazzler.config.static_folder
        )
        if os.path.exists(static_dir):
            self.app.router.add_static(
                f'{prefix}{self.dazzler.config.static_prefix}',
                static_dir
            )

        # Main internal static directory where requirements are copied.
        self.app.router.add_static(
            f'{prefix}/dazzler/requirements/static',
            self.dazzler.requirements_dir,
        )

    async def route_update(self, request: web.Request, page: Page):
        """
        WebSocket route for aspect updating.

        :param request: The incoming request.
        :param page: The incoming page.
        :return:
        """
        ws = web.WebSocketResponse()

        await ws.prepare(request)

        self.websockets.add(ws)

        request_queue = asyncio.Queue()
        pendings = []
        aspect_requests = {}
        storage_requests = {}
        done = asyncio.Event()

        async def request_loop():
            while not done.is_set():
                req = await request_queue.get()
                kind = req['kind']
                if kind == 'get-aspect':
                    aspect_requests[req['request_id']] = req.pop('queue')
                elif kind == 'get-storage':
                    storage_requests[req['request_id']] = req.pop('queue')
                await ws.send_json(req)

        def done_callback(done: asyncio.Task):
            pendings.remove(done)
            exception = done.exception()
            if exception and not done.cancelled():
                done.print_stack(file=sys.stderr)
                self.logger.error(exception)

        async def handler():
            async for msg in ws:  # type: aiohttp.WSMessage
                if msg.type == aiohttp.WSMsgType.TEXT:
                    # noinspection PyNoneFunctionAssignment
                    data = msg.json()
                    kind = data.get('kind')

                    if kind == 'binding':
                        binding = page.get_binding(data['key'])

                        task = self.loop.create_task(
                            binding(request, data, ws, request_queue)
                        )
                        pendings.append(task)
                        task.add_done_callback(done_callback)

                    elif kind == 'get-aspect':
                        request_id = data['request_id']
                        value = data.get('value', UNDEFINED)
                        req = aspect_requests.pop(request_id)

                        if value is UNDEFINED:
                            await req.put((UNDEFINED, data.get('error')))
                        else:
                            await req.put((value, UNDEFINED))

                    elif kind == 'get-storage':
                        request_id = data['request_id']
                        value = data.get('value', UNDEFINED)
                        queue = storage_requests.pop(request_id)
                        await queue.put((value, UNDEFINED))
                else:
                    self.logger.debug(f'No handler for msg type: {msg.type}')

            done.set()

        async def pong():
            while not done.is_set():
                await asyncio.sleep(self.dazzler.config.renderer.ping_interval)
                await ws.send_json({'kind': 'ping'})

        try:
            ops = [handler(), request_loop()]
            if self.dazzler.config.renderer.ping:
                ops.append(pong())
            await asyncio.gather(*ops)
        finally:
            for pending in pendings:
                pending.cancel()
            self.websockets.discard(ws)

        return ws

    async def route_page(self, _: web.Request, page: Page = None):
        """
        Index route for a page.

        :param _:
        :param page: The page to serve.
        :return:
        """
        script = {
            'src': '/dazzler/requirements/static/index.js',
            'data-retries': str(self.dazzler.config.renderer.retries),
            'id': 'dazzler-script'
        }
        if self.dazzler.config.renderer.ping:
            script.update({
                'data-ping': 'true',
                'data-ping-interval':
                    str(self.dazzler.config.renderer.ping_interval)
            })

        index = replace_all(
            self.index,
            page_title=page.title,
            renderer_scripts='\n'.join(
                x.tag(
                    dev=self.debug,
                    external=self.dazzler.config.requirements.prefer_external
                )
                for x in renderer.requirements if x.kind == 'js'
            ),
            css='\n'.join(
                x.tag(
                    dev=self.debug,
                    external=self.dazzler.config.requirements.prefer_external
                )
                for x in renderer.requirements if x.kind == 'css'
            ),
            dazzler_script=format_tag('script', script),
            meta='\n'.join(format_tag('meta', x) for x in page.meta_tags),
            favicon=format_tag(
                'link', {
                    'rel': 'icon',
                    'href': page.favicon,
                    'type': 'image/x-icon'
                }
            ) if page.favicon else '',
            header=page.header,
            footer=page.footer,
            lang=page.lang,
        )
        return web.Response(body=index, content_type='text/html')

    # noinspection PyMethodMayBeStatic
    async def route_page_json(self, request: web.Request, page: Page = None):
        """
        Serve the bindings and layout associated with page.

        :param request:
        :param page:
        :return:
        """
        prepared = await page.prepare(
            request,
            self.debug,
            external=self.dazzler.config.requirements.prefer_external
        )
        # Add global requirements first.
        prepared['requirements'] = [
            x.prepare(
                dev=self.debug,
                external=self.dazzler.config.requirements.prefer_external
            ) for x in self.dazzler.requirements
        ] + prepared['requirements']

        if self.dazzler.config.development.reload:
            # Reload mode needs the websocket even if no binding.
            prepared['reload'] = True

        return web.json_response(prepared)

    async def route_get_page(self, request: web.Request):
        page = request.query.get('page')
        if page in self.dazzler.pages:
            raise web.HTTPFound(location=request.app.router[page].url_for())
        raise web.HTTPNotFound()

    async def start(
            self, host: str, port: int,
            shutdown_timeout: float = 60.0,
            ssl_context: Optional[SSLContext] = None,
            backlog: int = 128,
            reuse_address: Optional[bool] = None,
            reuse_port: Optional[bool] = None,
    ):
        """
        Start the server

        :param host:
        :param port:
        :param shutdown_timeout:
        :param ssl_context:
        :param backlog:
        :param reuse_address:
        :param reuse_port:
        :return:
        """
        self.app.on_shutdown.append(self._on_shutdown)
        self.logger.debug('Starting server')
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        self.site = web.TCPSite(
            self.runner, host, port,
            shutdown_timeout=shutdown_timeout,
            ssl_context=ssl_context, backlog=backlog,
            reuse_address=reuse_address, reuse_port=reuse_port
        )
        await self.site.start()
        self.logger.info(f'Started server http://{host}:{port}/')

    async def _on_shutdown(self, _):
        # Close all websocket's and stop the main loop
        self.dazzler.stop_event.set()
        for ws in set(self.websockets):
            await ws.close(code=WSCloseCode.GOING_AWAY,
                           message='Server shutdown')

    async def send_reload(self, filenames, hot, refresh, deleted):
        for ws in set(self.websockets):
            await ws.send_json({
                'kind': 'reload',
                'filenames': filenames,
                'hot': hot,
                'refresh': refresh,
                'deleted': deleted
            })

    def _apply_middleware(self, handler):
        if not self.dazzler.middlewares:
            return handler

        async def apply(request: web.Request, *args, **kwargs):

            callbacks = []
            for middleware in self.dazzler.middlewares:
                callback = await middleware(request)
                if callback:
                    callbacks.append(callback)

            response = await handler(request, *args, **kwargs)

            for callback in callbacks:
                await callback(response)

            return response

        return apply



"""
Component packages of dazzler.

- ``core``: Lightweight interactive components.
- ``extra``: More components that may not be needed by all pages.
- ``markdown``: Markdown and CodeBlock.
- ``calendar``: Date & time related components.
"""



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Input(Component):
    """
    Html input wrapper.
    """
    _aspect_keys = []
    _children = []
    id = Aspect()
    class_name = Aspect()
    style = Aspect()
    name = Aspect()
    required = Aspect()
    value = Aspect(default='', docstring="(default='')")
    type = Aspect(docstring="(Possible values: 'hidden', 'text', 'number', 'search', 'tel', 'password', 'range', 'email', 'url', 'submit', 'checkbox', 'reset')")  # noqa: E501
    placeholder = Aspect()
    pattern = Aspect()
    auto_complete = Aspect()
    auto_focus = Aspect()
    auto_save = Aspect()
    auto_correct = Aspect()
    disabled = Aspect()
    n_blur = Aspect(default=0, docstring="(default=0)")
    n_submit = Aspect(default=0, docstring="(default=0)")
    min = Aspect()
    min_length = Aspect()
    max_length = Aspect()
    max = Aspect()
    step = Aspect()
    multiple = Aspect()
    checked = Aspect(default=False, docstring="(default=False)")

    def __init__(
            self,
            id: typing.Optional[str] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            name: typing.Optional[str] = UNDEFINED,
            required: typing.Optional[bool] = UNDEFINED,
            value: typing.Optional[typing.Union[str, typing.Union[float, int]]] = '',  # noqa: E501
            type: typing.Optional[typing.Any] = UNDEFINED,
            placeholder: typing.Optional[str] = UNDEFINED,
            pattern: typing.Optional[str] = UNDEFINED,
            auto_complete: typing.Optional[str] = UNDEFINED,
            auto_focus: typing.Optional[bool] = UNDEFINED,
            auto_save: typing.Optional[str] = UNDEFINED,
            auto_correct: typing.Optional[str] = UNDEFINED,
            disabled: typing.Optional[bool] = UNDEFINED,
            n_blur: typing.Optional[typing.Union[float, int]] = 0,
            n_submit: typing.Optional[typing.Union[float, int]] = 0,
            min: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            min_length: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            max_length: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            max: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            step: typing.Optional[typing.Union[str, typing.Union[float, int]]] = UNDEFINED,  # noqa: E501
            multiple: typing.Optional[bool] = UNDEFINED,
            checked: typing.Optional[bool] = False,
            identity: str = None
    ):
        """
        :param id:
        :param class_name:
        :param style:
        :param name:
        :param required:
        :param value: (default='')
        :param type: (Possible values: 'hidden', 'text', 'number',
            'search', 'tel', 'password', 'range', 'email', 'url', 'submit',
            'checkbox', 'reset')
        :param placeholder:
        :param pattern:
        :param auto_complete:
        :param auto_focus:
        :param auto_save:
        :param auto_correct:
        :param disabled:
        :param n_blur: (default=0)
        :param n_submit: (default=0)
        :param min:
        :param min_length:
        :param max_length:
        :param max:
        :param step:
        :param multiple:
        :param checked: (default=False)
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Form(Component):
    """
    A form element with auto fields.

    :CSS:

        ``dazzler-core-form``
        - ``form-header``
        - ``form-body``
        - ``form-field``
        - ``field-error``
        - ``form-label``
        - ``form-input``
        - ``form-footer``
        - ``form-submit``
    """
    _aspect_keys = []
    _children = []
    fields = Aspect(children=True, docstring="Fields of the form, either provide a component or a type for the input.")  # noqa: E501
    header = Aspect(children=True, docstring="Render on top of the form (CSS: form-header)")  # noqa: E501
    body = Aspect(children=True, docstring="Render in the middle of the form, after the fields. (CSS: form-body)")  # noqa: E501
    footer = Aspect(children=True, docstring="Render at the bottom of the form (CSS: form-footer)")  # noqa: E501
    action = Aspect(docstring="Url to submit the form.")
    method = Aspect(docstring="Http method to submit the form. (Possible values: 'get', 'post')")  # noqa: E501
    target = Aspect(docstring="How will the response display. (Possible values: '_blank', '_self', '_parent', '_top')")  # noqa: E501
    name = Aspect(docstring="Name of the form.")
    auto_complete = Aspect(docstring="Specifies auto complete for the form. (Possible values: 'on', 'off')")  # noqa: E501
    enctype = Aspect(docstring="How the data is encode before submit. (Possible values: 'application/x-www-form-urlencoded', 'multipart/form-data', 'text/plain')")  # noqa: E501
    accept_charset = Aspect(docstring="Character encoding for submission.")
    no_validate = Aspect(docstring="Specify the form should be validated on submission.")  # noqa: E501
    class_name = Aspect(docstring="CSS class of the form element.")
    style = Aspect(docstring="Style object of the form element.")
    errors = Aspect(default=UNDEFINED, docstring="Errors to show with the fields. Keys are name. (default={})")  # noqa: E501
    include_submit = Aspect(default=True, docstring="Include a submit button on the form. (default=True)")  # noqa: E501
    submit_label = Aspect(default='Submit', docstring="Label of the submit button. (default='Submit')")  # noqa: E501

    def __init__(
            self,
            fields: typing.Optional[typing.List[typing.Dict[str, typing.Union[str, typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]]]]] = UNDEFINED,  # noqa: E501
            header: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            body: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            footer: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            action: typing.Optional[str] = UNDEFINED,
            method: typing.Optional[typing.Any] = UNDEFINED,
            target: typing.Optional[typing.Any] = UNDEFINED,
            name: typing.Optional[str] = UNDEFINED,
            auto_complete: typing.Optional[typing.Any] = UNDEFINED,
            enctype: typing.Optional[typing.Any] = UNDEFINED,
            accept_charset: typing.Optional[str] = UNDEFINED,
            no_validate: typing.Optional[bool] = UNDEFINED,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            errors: typing.Optional[typing.Dict] = UNDEFINED,
            include_submit: typing.Optional[bool] = True,
            submit_label: typing.Optional[str] = 'Submit',
            identity: str = None
    ):
        """
        :param fields: Fields of the form, either provide a component or a
            type for the input.
        :param header: Render on top of the form (CSS: form-header)
        :param body: Render in the middle of the form, after the fields.
            (CSS: form-body)
        :param footer: Render at the bottom of the form (CSS: form-footer)
        :param action: Url to submit the form.
        :param method: Http method to submit the form. (Possible values:
            'get', 'post')
        :param target: How will the response display. (Possible values:
            '_blank', '_self', '_parent', '_top')
        :param name: Name of the form.
        :param auto_complete: Specifies auto complete for the form.
            (Possible values: 'on', 'off')
        :param enctype: How the data is encode before submit. (Possible
            values: 'application/x-www-form-urlencoded', 'multipart/form-data',
            'text/plain')
        :param accept_charset: Character encoding for submission.
        :param no_validate: Specify the form should be validated on
            submission.
        :param class_name: CSS class of the form element.
        :param style: Style object of the form element.
        :param errors: Errors to show with the fields. Keys are name.
            (default={})
        :param include_submit: Include a submit button on the form.
            (default=True)
        :param submit_label: Label of the submit button. (default='Submit')
        """
        Component.__init__(self, locals(), identity)



import os
import importlib


__all__ = [
    'get_package_path',
    'replace_all',
    'format_tag',
    'get_member',
]


def get_package_path(name):
    module = importlib.import_module(name)
    return os.path.dirname(module.__file__)


def get_member(member_path: str):
    """
    Import a module and retrieve a member from a module dot notation path
    separated by an ``:`` with the instance name.

    (eg: ``package.module:instance``)

    :param member_path: A string path with pattern
        ``package.module:member``
    :return:
    """
    split = member_path.split(':')
    if len(split) != 2:
        raise ImportError(
            'Invalid instance path syntax '
            '(eg: `package.module:instance`)'
        )
    module_name, instance_name = split
    module = importlib.import_module(module_name)
    return getattr(module, instance_name, None)


def replace_all(template, **kwargs):
    t = template
    for k, v in kwargs.items():
        t = t.replace(f'%({k})', str(v))
    return t


def format_tag(tag_name, attributes, content=None, opened=True, close=True):
    attrs = []
    for k, v in attributes.items():
        if isinstance(v, str):
            attrs.append(f'{k}="{v}"')
        else:
            fmt, value = v
            attrs.append(fmt.format(key=k, value=value))

    attr = ' '.join(attrs)
    tag = f'<{tag_name} {attr}'

    if opened:
        tag += '>'
    else:
        tag += '/>'

    if content:
        tag += content

    if close:
        tag += f'</{tag_name}>'

    return tag



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Interval(Component):
    """
    Update ``times`` aspect every interval to trigger a binding.
    """
    _aspect_keys = []
    _children = []
    timeout = Aspect(default=1000, docstring="The delay between each time. (default=1000)")  # noqa: E501
    times = Aspect(default=0, docstring="Number of times the interval was fired. (default=0)")  # noqa: E501
    active = Aspect(default=True, docstring="Enable/disable the interval loop. (default=True)")  # noqa: E501

    def __init__(
            self,
            timeout: typing.Optional[typing.Union[float, int]] = 1000,
            times: typing.Optional[typing.Union[float, int]] = 0,
            active: typing.Optional[bool] = True,
            identity: str = None
    ):
        """
        :param timeout: The delay between each time. (default=1000)
        :param times: Number of times the interval was fired. (default=0)
        :param active: Enable/disable the interval loop. (default=True)
        """
        Component.__init__(self, locals(), identity)



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Logout(Component):
    """
    A logout button.

    :CSS:

        ``dazzler-auth-logout``
        - ``logout-button``
    """
    _aspect_keys = []
    _children = []
    logout_url = Aspect(required=True, docstring="Logout url")
    next_url = Aspect(docstring="Redirect to this page after logout.")
    label = Aspect(default='Logout', docstring="Label of the logout button. (default='Logout')")  # noqa: E501
    method = Aspect(default='POST', docstring="Method to submit the logout form. (default='POST')")  # noqa: E501
    class_name = Aspect()
    style = Aspect()

    def __init__(
            self,
            logout_url: str,
            next_url: typing.Optional[str] = UNDEFINED,
            label: typing.Optional[str] = 'Logout',
            method: typing.Optional[str] = 'POST',
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param logout_url: Logout url
        :param next_url: Redirect to this page after logout.
        :param label: Label of the logout button. (default='Logout')
        :param method: Method to submit the logout form. (default='POST')
        :param class_name:
        :param style:
        """
        Component.__init__(self, locals(), identity)



import typing


class _Undefined(object):
    """
    Mark a difference between None and not supplied.
    The user can set a value to None and we should take it into account when
    serializing the component to include that aspect.
    A falsy check is not enough.
    """
    _singleton = None  # There can only be one, like None
    __slots__ = ()  # no attribute can be added.

    @classmethod  # important, otherwise not the good __new__
    def __new__(cls, *args, **kwargs):
        if not cls._singleton:
            cls._singleton = super(_Undefined, cls).__new__(*args, **kwargs)
            return cls._singleton
        raise Exception('There can only be one UNDEFINED')

    def __repr__(self):
        return 'undefined'

    def __str__(self):
        return 'undefined'

    def __bool__(self):
        return False


# Includes None for non required aspects to be set to None/null
Undefined = typing.Union[_Undefined, None]

UNDEFINED = _Undefined()



"""Autogenerated file: DO NOT EDIT!"""
import typing  # noqa: F401
from dazzler.system import Component, Aspect, UNDEFINED  # noqa: F401


class Image(Component):
    """
    An image.

    :CSS: ``dazzler-core-image``
    """
    _aspect_keys = []
    _children = []
    src = Aspect(required=True, docstring="The source url of the image.")
    alt = Aspect(required=True, docstring="Alt img attribute to show when the browser cannot display the image.")  # noqa: E501
    height = Aspect(docstring="Height in pixel")
    width = Aspect(docstring="Width in pixel")
    link = Aspect(docstring="Link to another page when the image is clicked.")
    caption = Aspect(children=True, docstring="Text to include beneath the image.")  # noqa: E501
    cross_origin = Aspect(docstring="Fetch the image with CORS. (Possible values: 'anonymous', 'use-credentials')")  # noqa: E501
    decoding = Aspect(docstring="Decoding hint for the browser (Possible values: 'sync', 'async', 'auto')")  # noqa: E501
    preload = Aspect(docstring="Preload the image before mount.")
    clicks = Aspect(default=0, docstring="Times the image was clicked on. (default=0)")  # noqa: E501
    class_name = Aspect(docstring="CSS class to give to root element, (Scoped: dazzler-core-image)")  # noqa: E501
    style = Aspect(docstring="Style object of the root element.")
    bordered = Aspect(docstring="Add bordered style class")
    rounded = Aspect(docstring="Add rounded style class")
    centered = Aspect(docstring="Add centered style class for the image toappear in the horizontal middle")  # noqa: E501
    flexible = Aspect(docstring="Take up the whole size of the parent container.")  # noqa: E501
    circle = Aspect(docstring="Image is a circle")
    padded = Aspect(docstring="Add a little padding around the image.")
    caption_style = Aspect(docstring="Style to give to the caption of the image.")  # noqa: E501
    caption_class_name = Aspect(docstring="CSS class to give to the caption of the image.")  # noqa: E501
    figure_class_name = Aspect(docstring="CSS class to give to figure if caption is provided.")  # noqa: E501
    figure_style = Aspect(docstring="Style object to give to figure if caption is provided.")  # noqa: E501
    link_class_name = Aspect(docstring="CSS class to give to the link element.")  # noqa: E501
    link_style = Aspect(docstring="Style object to give to the link.")

    def __init__(
            self,
            src: str,
            alt: str,
            height: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            width: typing.Optional[typing.Union[float, int]] = UNDEFINED,
            link: typing.Optional[str] = UNDEFINED,
            caption: typing.Optional[typing.Union[str, int, float, Component, typing.List[typing.Union[str, int, float, Component]]]] = UNDEFINED,  # noqa: E501
            cross_origin: typing.Optional[typing.Any] = UNDEFINED,
            decoding: typing.Optional[typing.Any] = UNDEFINED,
            preload: typing.Optional[bool] = UNDEFINED,
            clicks: typing.Optional[typing.Union[float, int]] = 0,
            class_name: typing.Optional[str] = UNDEFINED,
            style: typing.Optional[typing.Dict] = UNDEFINED,
            bordered: typing.Optional[bool] = UNDEFINED,
            rounded: typing.Optional[bool] = UNDEFINED,
            centered: typing.Optional[bool] = UNDEFINED,
            flexible: typing.Optional[bool] = UNDEFINED,
            circle: typing.Optional[bool] = UNDEFINED,
            padded: typing.Optional[bool] = UNDEFINED,
            caption_style: typing.Optional[typing.Dict] = UNDEFINED,
            caption_class_name: typing.Optional[str] = UNDEFINED,
            figure_class_name: typing.Optional[str] = UNDEFINED,
            figure_style: typing.Optional[typing.Dict] = UNDEFINED,
            link_class_name: typing.Optional[str] = UNDEFINED,
            link_style: typing.Optional[typing.Dict] = UNDEFINED,
            identity: str = None
    ):
        """
        :param src: The source url of the image.
        :param alt: Alt img attribute to show when the browser cannot
            display the image.
        :param height: Height in pixel
        :param width: Width in pixel
        :param link: Link to another page when the image is clicked.
        :param caption: Text to include beneath the image.
        :param cross_origin: Fetch the image with CORS. (Possible values:
            'anonymous', 'use-credentials')
        :param decoding: Decoding hint for the browser (Possible values:
            'sync', 'async', 'auto')
        :param preload: Preload the image before mount.
        :param clicks: Times the image was clicked on. (default=0)
        :param class_name: CSS class to give to root element, (Scoped:
            dazzler-core-image)
        :param style: Style object of the root element.
        :param bordered: Add bordered style class
        :param rounded: Add rounded style class
        :param centered: Add centered style class for the image to appear
            in the horizontal middle
        :param flexible: Take up the whole size of the parent container.
        :param circle: Image is a circle
        :param padded: Add a little padding around the image.
        :param caption_style: Style to give to the caption of the image.
        :param caption_class_name: CSS class to give to the caption of the
            image.
        :param figure_class_name: CSS class to give to figure if caption is
            provided.
        :param figure_style: Style object to give to figure if caption is
            provided.
        :param link_class_name: CSS class to give to the link element.
        :param link_style: Style object to give to the link.
        """
        Component.__init__(self, locals(), identity)
